@@ -23,14 +23,16 @@ from PySide6 import QtCore
 from PySide6.QtGui import (QPixmap, QIcon, QCursor)
 from PySide6.QtWidgets import (QPushButton, QApplication,
                                QVBoxLayout, QHBoxLayout, QDialog, QLabel,
-                               QMessageBox, QComboBox, QProgressBar)
+                               QMessageBox, QComboBox, QProgressBar,
+                               QCheckBox, QFormLayout)
 from qt_material import apply_stylesheet
 
+from meshtastic_flasher.version import __version__
+
 # windows does not like this one
 if platform.system() != "Windows":
     import grp
 
-from meshtastic_flasher.version import __version__
 
 MESHTASTIC_LOGO_FILENAME = "logo.png"
 MESHTASTIC_COLOR_DARK = "#2C2D3C"
@@ -56,6 +58,38 @@ def populate_tag_in_firmware_dropdown(tag):
     print(f'tag:{tag} populate in dropdown?:{retval}')
     return retval
 
+class AdvancedForm(QDialog):
+    """Advanced options form"""
+
+    def __init__(self, parent=None):
+        """constructor"""
+        super(AdvancedForm, self).__init__(parent)
+
+        width = 200
+        height = 100
+        self.setMinimumSize(width, height)
+        self.setWindowTitle("Advanced Options")
+
+        # Create widgets
+        self.update_only_cb = QCheckBox()
+        self.update_only_cb.setToolTip("If enabled, the device will be updated (not completely erased).")
+
+        self.ok_button = QPushButton("OK")
+
+        # create form
+        form_layout = QFormLayout()
+        form_layout.addRow(self.tr("&Update only"), self.update_only_cb)
+        form_layout.addRow(self.tr(""), self.ok_button)
+        self.setLayout(form_layout)
+
+        self.ok_button.clicked.connect(self.close_advanced_options)
+
+    def close_advanced_options(self):
+        """Test if works"""
+        print('OK button was clicked in advanced options')
+        self.close()
+
+
 class Form(QDialog):
     """Main application"""
 
@@ -70,16 +104,18 @@ class Form(QDialog):
         self.nrf = False
         self.device = None
 
+        self.advanced_form = AdvancedForm()
+
         self.setWindowTitle(f"Meshtastic Flasher v{__version__}")
 
         # Create widgets
-        self.select_firmware = QPushButton("SELECT FIRMWARE")
-        self.select_firmware.setToolTip("Click to check for more recent firmware.")
+        self.get_versions = QPushButton("GET VERSIONS")
+        self.get_versions.setToolTip("Click to check for more recent firmware.")
 
         self.select_firmware_version = QComboBox()
-        self.select_firmware_version.setToolTip("Select which firmware to flash.")
+        self.select_firmware_version.setToolTip("Click GET VERSIONS for the list of firmware versions.")
         self.select_firmware_version.setMinimumContentsLength(18)
-        self.select_firmware_version.hide()
+        self.select_firmware_version.setDisabled(True)
 
         self.select_detect = QPushButton("DETECT DEVICE")
         self.select_detect.setToolTip("Click to detect supported device and port info.")
@@ -87,17 +123,17 @@ class Form(QDialog):
         self.select_detect.setStyleSheet("text-transform: none")
 
         self.select_port = QComboBox()
-        self.select_port.setToolTip("Click SELECT FIRMWARE before you can select the port.")
+        self.select_port.setToolTip("Click GET VERSIONS and DETECT DEVICE before you can select the port.")
         self.select_port.setMinimumContentsLength(25)
         self.select_port.setDisabled(True)
 
         self.select_device = QComboBox()
-        self.select_device.setToolTip("Click SELECT FIRMWARE before you can select the device.")
+        self.select_device.setToolTip("Click GET VERSIONS and DETECT DEVICE before you can select the device.")
         self.select_device.setMinimumContentsLength(17)
         self.select_device.setDisabled(True)
 
         self.select_flash = QPushButton("FLASH")
-        self.select_flash.setToolTip("Click to flash the firmware. If button is not enabled, need to click the buttons to the left.")
+        self.select_flash.setToolTip("Click to flash the firmware.\nIf this button is not enabled, need to click the GET VERSIONS and DETECT DEVICE \nbuttons to populate the available options.")
         self.select_flash.setEnabled(False)
 
         self.progress = QProgressBar()
@@ -114,44 +150,89 @@ class Form(QDialog):
         self.logo.setStyleSheet(style_for_logo)
         self.logo.setCursor(QCursor(QtCore.Qt.PointingHandCursor))
 
+        # labels for over the drop downs/combo boxes
+        self.label_version = QLabel(self)
+        self.label_version.setText("                         Version")
+
+        self.label_port = QLabel(self)
+        self.label_port.setText("                           Port")
+
+        self.label_device = QLabel(self)
+        self.label_device.setText("                        Device")
+
         # Create layout and add widgets
         main_layout = QVBoxLayout()
         main_layout.setContentsMargins(0, 0, 0, 0)
         if self.logo:
             main_layout.addWidget(self.logo)
+
         main_layout.addStretch(1)
+
+        detect_layout = QHBoxLayout()
+        detect_layout.addStretch(1)
+        detect_layout.addWidget(self.get_versions)
+        detect_layout.addWidget(self.select_detect)
+        detect_layout.setContentsMargins(0, 0, 0, 0)
+        detect_layout.addStretch(1)
+
+        label_layout = QHBoxLayout()
+        label_layout.addWidget(self.label_version)
+        label_layout.addWidget(self.label_port)
+        label_layout.addWidget(self.label_device)
+        label_layout.setContentsMargins(0, 0, 0, 0)
+
         button_layout = QHBoxLayout()
         button_layout.addStretch(1)
-        button_layout.addWidget(self.select_firmware)
         button_layout.addWidget(self.select_firmware_version)
-        button_layout.addWidget(self.select_detect)
         button_layout.addWidget(self.select_port)
         button_layout.addWidget(self.select_device)
-        button_layout.addWidget(self.select_flash)
         button_layout.addStretch(1)
-        button_layout.setContentsMargins(20, 20, 20, 20)
+        button_layout.setContentsMargins(20, 0, 20, 0)
+
+        flash_layout = QHBoxLayout()
+        flash_layout.addStretch(1)
+        flash_layout.addWidget(self.select_flash)
+        flash_layout.addStretch(1)
+        flash_layout.setContentsMargins(20, 0, 20, 20)
 
         progress_layout = QHBoxLayout()
         progress_layout.addWidget(self.progress)
         button_layout.addStretch(1)
 
         # Set layout
+        main_layout.addLayout(detect_layout)
+        main_layout.addLayout(label_layout)
         main_layout.addLayout(button_layout)
+        main_layout.addLayout(flash_layout)
         main_layout.addLayout(progress_layout)
         main_layout.addStretch(1)
         self.setLayout(main_layout)
 
         # Add button signals to slots
         self.logo.mousePressEvent = self.logo_clicked
-        self.select_firmware.clicked.connect(self.download_firmware_versions)
+        self.get_versions.clicked.connect(self.get_versions_from_github)
         self.select_detect.clicked.connect(self.detect)
         self.select_flash.clicked.connect(self.flash_stuff)
         self.select_firmware_version.currentTextChanged.connect(self.on_select_firmware_changed)
 
+    def keyPressEvent(self, event):
+        """Deal with a key press"""
+        if event.key() == QtCore.Qt.Key_A:
+            print("A was pressed... showing advanced options form")
+            self.show_advanced_options()
+        elif event.key() == QtCore.Qt.Key_Q:
+            print("Q was pressed... so quitting")
+            QApplication.quit()
+
 
     def on_select_firmware_changed(self, value):
         """When the select_firmware drop down value is changed."""
         print(f'on_select_firmware_changed value:{value}')
+
+        QApplication.processEvents()
+        self.progress.setValue(0)
+        self.progress.show()
+
         self.firmware_version = self.select_firmware_version.currentText()[1:] # drop leading v
 
         # zip filename from tag
@@ -167,6 +248,9 @@ class Form(QDialog):
         if not os.path.exists(zip_file_name):
             print("Need to download...")
 
+            self.progress.setValue(20)
+            QApplication.processEvents()
+
             # Note: Probably should use the browser_download_url
             zip_file_url = f'https://github.com/meshtastic/Meshtastic-device/releases/download/v{self.firmware_version}/firmware-{self.firmware_version}.zip'
             # This is in case we have to temp-disable GitHub during dev due to rate-limiting.
@@ -182,20 +266,25 @@ class Form(QDialog):
             urllib.request.urlretrieve(zip_file_url, zip_file_name)
             print("done downloading")
 
+            self.progress.setValue(80)
+            QApplication.processEvents()
+
         # unzip into directory named the same name as the firmware_version
         if not os.path.exists(self.firmware_version):
             print("Unzipping files now...")
             with zipfile.ZipFile(zip_file_name, 'r') as zip_ref:
                 zip_ref.extractall(self.firmware_version)
             print("done unzipping")
+        self.progress.setValue(100)
+        QApplication.processEvents()
 
-    def download_firmware_versions(self):
+    def get_versions_from_github(self):
         """Download versions from GitHub"""
 
         # save first_tag in case we need to populate list with *some* value
         first_tag = None
 
-        if not self.select_firmware.isHidden():
+        if not self.select_firmware_version.count() == 0:
 
             try:
                 token = Github()
@@ -235,9 +324,8 @@ class Form(QDialog):
                 fall_back_tag = 'v1.2.53.19c1f9f'
                 self.select_firmware_version.addItem(fall_back_tag)
 
-        # if we checked for latest versions, so hide the "Firmware" button, and show the combo box to select version
-        self.select_firmware.hide()
-        self.select_firmware_version.show()
+        self.select_firmware_version.setEnabled(True)
+        self.select_firmware_version.setToolTip("Select desired firmware version to flash.")
 
         # only enable Flash button if we have both values
         if self.select_port.count() > 0 and self.firmware_version:
@@ -250,6 +338,7 @@ class Form(QDialog):
                 self.select_device.insertSeparator(self.select_device.count())
                 self.select_device.addItem('All')
                 count = self.select_device.count() - 1
+                # not make the label 'All' selectable
                 self.select_device.model().item(count).setEnabled(False)
                 filenames = next(os.walk(self.firmware_version), (None, None, []))[2]
                 filenames.sort()
@@ -260,10 +349,18 @@ class Form(QDialog):
                         device = device.replace(".bin", "")
                         self.select_device.addItem(device)
 
+    def show_advanced_options(self):
+        """Advanced Options"""
+        print("advanced options")
+        self.advanced_form.show()
 
     def detect(self):
         """Detect port, download zip file from github if we need to, and unzip it"""
 
+        QApplication.processEvents()
+        self.progress.setValue(0)
+        self.progress.show()
+
         system = platform.system()
         if system == 'Linux':
             username = os.getlogin()
@@ -279,6 +376,9 @@ class Form(QDialog):
                 dlg.setText(message)
                 dlg.exec()
 
+        self.progress.setValue(10)
+        QApplication.processEvents()
+
         # detect supported devices
         supported_devices_detected = detect_supported_devices()
         if len(supported_devices_detected) == 0:
@@ -290,6 +390,7 @@ class Form(QDialog):
             if len(supported_devices_detected) > 0:
                 self.select_device.clear()
                 self.select_device.addItem('Detected')
+                # not make the label 'Detected' selectable
                 self.select_device.model().item(0).setEnabled(False)
                 for device in supported_devices_detected:
                     print(f'Detected {device.name}')
@@ -297,6 +398,9 @@ class Form(QDialog):
                 if self.select_device.count() > 1:
                     self.select_device.setCurrentIndex(1)
 
+        self.progress.setValue(40)
+        QApplication.processEvents()
+
         # detect which ports and populate the dropdown
         ports = active_ports_on_supported_devices(supported_devices_detected)
         ports_sorted = list(ports)
@@ -307,6 +411,9 @@ class Form(QDialog):
                 possible_weird = True
             self.select_port.addItem(port)
 
+        self.progress.setValue(70)
+        QApplication.processEvents()
+
         if possible_weird:
             # deal with weird TLora (single device connected, but shows up as 2 ports)
             # ports:['/dev/cu.usbmodem533C0052151', '/dev/cu.wchusbserial533C0052151']
@@ -318,10 +425,13 @@ class Form(QDialog):
                 if first == second:
                     print('We are dealing with a weird TLora port situation.')
                     self.select_port.clear()
-                    self.select_port.addItem(ports[1])
-                    self.select_port.addItem(ports[0]) # delete this one?
+                    self.select_port.addItem(tmp_ports[1])
+                    self.select_port.addItem(tmp_ports[0]) # delete this one?
                     ports = tmp_ports
 
+        self.progress.setValue(80)
+        QApplication.processEvents()
+
         # our auto-detect did not work
         if len(ports) == 0:
             print("Warning: Could not find any ports using the autodetection method.")
@@ -337,6 +447,9 @@ class Form(QDialog):
                 for port in ports:
                     self.select_port.addItem(port)
 
+        self.progress.setValue(90)
+        QApplication.processEvents()
+
         self.all_devices()
 
         # NRF52 devices
@@ -350,32 +463,34 @@ class Form(QDialog):
 
             partitions = psutil.disk_partitions()
             #print(f'partitions:{partitions}')
-            search_for_partition = 'FTHR840BOOT'
+            search_for_partition = ['FTHR840BOOT', 'TECHOBOOT']
             found_partition = False
 
             if platform.system() == "Windows":
                 # need to run some power shell
-                _, gv_output = subprocess.getstatusoutput('powershell.exe "Get-Volume"')
-                if re.search(search_for_partition, str(gv_output), re.MULTILINE):
-                    print('found partition on windows!')
-                    found_partition = True
-                    # for each line of output
-                    lines = str(gv_output).split('\n')
-                    print(f'lines:{lines}')
-                    for line in lines:
-                        parts = line.split(' ')
-                        print(f'parts:{parts}')
-                        if search_for_partition in line:
-                            self.select_port.addItem(f"{parts[0]}:/")
-                            break
+                _, gv_output = subprocess.getstatusoutput('powershell.exe "[Console]::OutputEncoding = [Text.UTF8Encoding]::UTF8; Get-Volume"')
+                for search in search_for_partition:
+                    if re.search(search, str(gv_output), re.MULTILINE):
+                        print('found partition on windows!')
+                        found_partition = True
+                        # for each line of output
+                        lines = str(gv_output).split('\n')
+                        print(f'lines:{lines}')
+                        for line in lines:
+                            parts = line.split(' ')
+                            #print(f'parts:{parts}')
+                            if search in line:
+                                self.select_port.addItem(f"{parts[0]}:/")
+                                break
             else: # Linux or Darwin
                 for partition in partitions:
                     print(f'partition:{partition}')
-                    if search_for_partition in partition.mountpoint:
-                        print('*** found search_for_partition')
-                        self.select_port.addItem(partition.mountpoint)
-                        found_partition = True
-                        break
+                    for search in search_for_partition:
+                        if search in partition.mountpoint:
+                            print(f'*** found {search}')
+                            self.select_port.addItem(partition.mountpoint)
+                            found_partition = True
+                            break
 
             if found_partition:
                 # the 19003 reports same as 5005, so we cannot really trust it
@@ -383,6 +498,7 @@ class Form(QDialog):
                 # and the T-Echo
                 self.select_device.clear()
                 self.select_device.addItem('Detected')
+                # do not make the label 'Detected' selectable
                 self.select_device.model().item(0).setEnabled(False)
                 self.select_device.addItem('rak4631_5005')
                 self.select_device.addItem('rak4631_19003')
@@ -400,14 +516,15 @@ class Form(QDialog):
 
         # only enable Flash button and Device dropdown if we have firmware and ports
         if self.select_port.count() > 0 and self.firmware_version:
-            self.select_port.setToolTip("Select the port.")
-            self.select_device.setToolTip("Select the device.")
+            self.select_port.setToolTip("Select the communication device")
+            self.select_device.setToolTip("Select the device variant")
             self.select_flash.setEnabled(True)
             self.select_flash.setToolTip('Click the FLASH button to write to the device.')
-            self.select_detect.hide()
             self.select_port.setDisabled(False)
             self.select_device.setDisabled(False)
 
+        self.progress.setValue(100)
+        QApplication.processEvents()
 
     def logo_clicked(self, event):
         """The logo was clicked."""
@@ -418,13 +535,25 @@ class Form(QDialog):
         """Do the flash parts"""
         proceed = False
 
+        QApplication.processEvents()
+        self.progress.setValue(0)
+        self.progress.show()
+
+        update_only = self.advanced_form.update_only_cb.isChecked()
+        update_only_message = ""
+        if update_only:
+            print('update only is checked')
+            update_only_message = "**update only** "
+        else:
+            print('update only is not checked')
+
         self.port = self.select_port.currentText()
         self.device = self.select_device.currentText()
 
         verb = 'flash'
         if self.nrf:
             verb = 'copy'
-        confirm_msg = f'Are you sure you want to {verb}\n{self.firmware_version}\n'
+        confirm_msg = f'Are you sure you want to {update_only_message}{verb}\n{self.firmware_version}\n'
         confirm_msg += f'{self.port}\n{self.device}?'
         reply = QMessageBox.question(self, 'Flash', confirm_msg,
                                      QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
@@ -447,34 +576,48 @@ class Form(QDialog):
 
             else:
                 # esp32 devices
-                QApplication.processEvents()
-                self.progress.show()
-                command = ["--baud", self.speed, "--port", self.port, "erase_flash"]
-                print(f"ESPTOOL Using command:{' '.join(command)}")
-                esptool.main(command)
-                self.progress.setValue(25)
-                QApplication.processEvents()
 
-                system_info_file = f"{self.firmware_version}/system-info.bin"
-                command = ["--baud", self.speed, "--port", self.port, "write_flash", "0x1000", system_info_file]
-                print(f"ESPTOOL Using command:{' '.join(command)}")
-                esptool.main(command)
-                self.progress.setValue(50)
-                QApplication.processEvents()
+                if update_only:
+                    device_file = f"{self.firmware_version}/firmware-{self.device}-{self.firmware_version}.bin"
+                    command = ["--baud", self.speed, "--port", self.port, "write_flash", "0x10000", device_file]
+                    print(f"ESPTOOL Using command:{' '.join(command)}")
+                    esptool.main(command)
+                    self.progress.setValue(50)
+                    QApplication.processEvents()
 
-                bin_file = f"{self.firmware_version}/spiffs-{self.firmware_version}.bin"
-                command = ["--baud", self.speed, "--port", self.port, "write_flash", "0x00390000", bin_file]
-                print(f"ESPTOOL Using command:{' '.join(command)}")
-                esptool.main(command)
-                self.progress.setValue(75)
-                QApplication.processEvents()
+                    command = ["--baud", self.speed, "--port", self.port, "erase_region", "0xe000", "0x2000"]
+                    print(f"ESPTOOL Using command:{' '.join(command)}")
+                    esptool.main(command)
+                    self.progress.setValue(100)
+                    QApplication.processEvents()
 
-                device_file = f"{self.firmware_version}/firmware-{self.device}-{self.firmware_version}.bin"
-                command = ["--baud", self.speed, "--port", self.port, "write_flash", "0x10000", device_file]
-                print(f"ESPTOOL Using command:{' '.join(command)}")
-                esptool.main(command)
-                self.progress.setValue(100)
-                QApplication.processEvents()
+                else:
+                    command = ["--baud", self.speed, "--port", self.port, "erase_flash"]
+                    print(f"ESPTOOL Using command:{' '.join(command)}")
+                    esptool.main(command)
+                    self.progress.setValue(25)
+                    QApplication.processEvents()
+
+                    system_info_file = f"{self.firmware_version}/system-info.bin"
+                    command = ["--baud", self.speed, "--port", self.port, "write_flash", "0x1000", system_info_file]
+                    print(f"ESPTOOL Using command:{' '.join(command)}")
+                    esptool.main(command)
+                    self.progress.setValue(50)
+                    QApplication.processEvents()
+
+                    bin_file = f"{self.firmware_version}/spiffs-{self.firmware_version}.bin"
+                    command = ["--baud", self.speed, "--port", self.port, "write_flash", "0x00390000", bin_file]
+                    print(f"ESPTOOL Using command:{' '.join(command)}")
+                    esptool.main(command)
+                    self.progress.setValue(75)
+                    QApplication.processEvents()
+
+                    device_file = f"{self.firmware_version}/firmware-{self.device}-{self.firmware_version}.bin"
+                    command = ["--baud", self.speed, "--port", self.port, "write_flash", "0x10000", device_file]
+                    print(f"ESPTOOL Using command:{' '.join(command)}")
+                    esptool.main(command)
+                    self.progress.setValue(100)
+                    QApplication.processEvents()
 
                 dlg2 = QMessageBox(self)
                 dlg2.setStyleSheet(f"background-color: {MESHTASTIC_COLOR_DARK}")
